open Datatypes
open Specif

type 'a list =
  | Coq_nil
  | Coq_cons of 'a * 'a list

(** val list_rect : 'a2 -> ('a1 -> 'a1 list -> 'a2 -> 'a2) -> 'a1 list -> 'a2 **)

let rec list_rect f f0 = function
  | Coq_nil -> f
  | Coq_cons (a, l0) -> f0 a l0 (list_rect f f0 l0)

(** val list_rec : 'a2 -> ('a1 -> 'a1 list -> 'a2 -> 'a2) -> 'a1 list -> 'a2 **)

let rec list_rec f f0 = function
  | Coq_nil -> f
  | Coq_cons (a, l0) -> f0 a l0 (list_rec f f0 l0)

(** val app : 'a1 list -> 'a1 list -> 'a1 list **)

let rec app l m =
  match l with
    | Coq_nil -> m
    | Coq_cons (a, l1) -> Coq_cons (a, (app l1 m))

(** val head : 'a1 list -> 'a1 option **)

let head = function
  | Coq_nil -> None
  | Coq_cons (x, l0) -> value x

(** val tail : 'a1 list -> 'a1 list **)

let tail = function
  | Coq_nil -> Coq_nil
  | Coq_cons (a, m) -> m

(** val length : 'a1 list -> int **)

let rec length = function
  | Coq_nil -> 0
  | Coq_cons (a, m) -> succ (length m)

(** val coq_In_dec : ('a1 -> 'a1 -> bool) -> 'a1 -> 'a1 list -> bool **)

let rec coq_In_dec h a = function
  | Coq_nil -> false
  | Coq_cons (a0, l0) ->
      (match h a0 a with
         | true -> true
         | false -> coq_In_dec h a l0)

(** val nth : int -> 'a1 list -> 'a1 -> 'a1 **)

let rec nth = (let rec nth n l default =
    assert(n >= 0);
    match n with
    | 0 -> (match l with
            | Coq_nil -> default
            | Coq_cons (x, l') -> x)
    | _ -> (match l with
            | Coq_nil -> default
            | Coq_cons (x, t) -> nth (pred n) t default) in nth)

(** val nth_ok : int -> 'a1 list -> 'a1 -> bool **)

let rec nth_ok = (let rec nth_ok n l default =
    assert(n >= 0);
    match n with
    | 0 -> (match l with
            | Coq_nil -> false
            | Coq_cons (x, l') -> true)
    | _ -> (match l with
            | Coq_nil -> false
            | Coq_cons (x, t) -> nth_ok (pred n) t default) in nth_ok)

(** val nth_in_or_default : int -> 'a1 list -> 'a1 -> bool **)

let rec nth_in_or_default = (let rec nth_in_or_default n l d =
  match l with
    | Coq_nil -> false
    | Coq_cons (a, l0) ->
        (match n with
           | 0 -> true
           | _ -> nth_in_or_default (pred n) l0 d) in  nth_in_or_default)

(** val nth_error : 'a1 list -> int -> 'a1 coq_Exc **)

let rec nth_error = (let rec nth_error l n =
  assert(n >= 0);
  match n with
  | 0 -> (match l with
            | Coq_nil -> None
            | Coq_cons (x, l0) -> value x)
  | _ ->
      (match l with
         | Coq_nil -> None
         | Coq_cons (a, l0) -> nth_error l0 (pred n)) in nth_error)

(** val nth_default : 'a1 -> 'a1 list -> int -> 'a1 **)

let nth_default default l n =
  match nth_error l n with
    | Some x -> x
    | None -> default

(** val list_eq_dec : ('a1 -> 'a1 -> bool) -> 'a1 list -> 'a1 list -> bool **)

let rec list_eq_dec h x y =
  match x with
    | Coq_nil ->
        (match y with
           | Coq_nil -> true
           | Coq_cons (a0, l0) -> false)
    | Coq_cons (a, l) ->
        (match y with
           | Coq_nil -> false
           | Coq_cons (a0, l0) ->
               (match h a a0 with
                  | true -> list_eq_dec h l l0
                  | false -> false))

(** val rev : 'a1 list -> 'a1 list **)

let rec rev = function
  | Coq_nil -> Coq_nil
  | Coq_cons (x, l') -> app (rev l') (Coq_cons (x, Coq_nil))

(** val map : ('a1 -> 'a2) -> 'a1 list -> 'a2 list **)

let rec map f = function
  | Coq_nil -> Coq_nil
  | Coq_cons (a, t) -> Coq_cons ((f a), (map f t))

(** val flat_map : ('a1 -> 'a2 list) -> 'a1 list -> 'a2 list **)

let rec flat_map f = function
  | Coq_nil -> Coq_nil
  | Coq_cons (x, t) -> app (f x) (flat_map f t)

(** val list_prod : 'a1 list -> 'a2 list -> ('a1, 'a2) pair list **)

let rec list_prod l l' =
  match l with
    | Coq_nil -> Coq_nil
    | Coq_cons (x, t) ->
        app
          (let rec map0 = function
             | Coq_nil -> Coq_nil
             | Coq_cons (a, t0) -> Coq_cons (( (x, a)), (map0 t0))
           in map0 l') (list_prod t l')

(** val list_power : 'a1 list -> 'a2 list -> ('a1, 'a2) pair list list **)

let rec list_power l l' =
  match l with
    | Coq_nil -> Coq_cons (Coq_nil, Coq_nil)
    | Coq_cons (x, t) ->
        flat_map (fun f ->
          let rec map0 = function
            | Coq_nil -> Coq_nil
            | Coq_cons (a, t0) -> Coq_cons ((Coq_cons (( (x, a)), f)),
                (map0 t0))
          in map0 l') (list_power t l')

(** val fold_left : ('a1 -> 'a2 -> 'a1) -> 'a2 list -> 'a1 -> 'a1 **)

let rec fold_left f l a0 =
  match l with
    | Coq_nil -> a0
    | Coq_cons (b, t) -> fold_left f t (f a0 b)

(** val fold_right : ('a2 -> 'a1 -> 'a1) -> 'a1 -> 'a2 list -> 'a1 **)

let rec fold_right f a0 = function
  | Coq_nil -> a0
  | Coq_cons (b, t) -> f b (fold_right f a0 t)

