Module Type ORDERED_SET.
Parameter domain : Set.
Axiom domain_le : domain -> domain -> Prop.
Axiom domain_le_refl : forall x : domain, (domain_le x x).
Axiom domain_le_trans : forall x y z : domain,
  (domain_le x y) -> (domain_le y z) -> (domain_le x z).

Inductive domain_comparison (x y : domain) : Set :=
  LESS : (domain_le x y) -> ~(domain_le y x) -> (domain_comparison x y)
| EQUAL : (domain_le x y) -> (domain_le y x) -> (domain_comparison x y)
| GREATER : (domain_le y x) -> ~(domain_le x y) -> (domain_comparison x y).

Parameter compare : forall x y : domain, (domain_comparison x y).
End ORDERED_SET.

Module Make (Domain : ORDERED_SET).
Import Domain.


Hint Immediate domain_le_trans : map2.
Hint Resolve domain_le_refl : map2.

Lemma le_then_LESS : forall x y : domain,
  (domain_le x y) ->
  ~(domain_le y x) ->
  forall S : Set, forall a b c : S,
  (match (compare x y) with
   | LESS _ _ => a
   | EQUAL _ _ => b
   | GREATER _ _ => c
   end) = a.
Proof.
  intros.
  destruct (compare x y); [ trivial | contradiction | contradiction ].
Qed.

Lemma le_then_EQUAL : forall x y : domain,
  (domain_le x y) ->
  (domain_le y x) ->
  forall S : Set, forall a b c : S,
  (match (compare x y) with
   | LESS _ _ => a
   | EQUAL _ _ => b
   | GREATER _ _ => c
   end) = b.
Proof.
  intros.
  destruct (compare x y); [ contradiction | trivial | contradiction ].
Qed.

Lemma le_then_GREATER : forall x y : domain,
  (domain_le y x) ->
  ~(domain_le x y) ->
  forall S : Set, forall a b c : S,
  (match (compare x y) with
   | LESS _ _ => a
   | EQUAL _ _ => b
   | GREATER _ _ => c
   end) = c.
Proof.
  intros.
  destruct (compare x y); [ contradiction | contradiction | trivial].
Qed.

Lemma le_refl_EQUAL : forall x : domain,
  forall S : Set, forall a b c : S,
  (match (compare x x) with
   | LESS _ _ => a
   | EQUAL _ _ => b
   | GREATER _ _ => c
   end) = b.
Proof.
  intros.
  apply le_then_EQUAL; auto with map2.
Qed.

Ltac solve_map2 := solve
         [ apply False_ind; eauto with map2 | eauto with map2 ] ||
          idtac.

Ltac compare_cases x y := destruct (compare x y); simpl; auto.

Require Import Arith.
Require Import Max.

Section Map2_set.
  Variable codomain : Set.

  Inductive map2_set : Set :=
    Empty_map2 : map2_set
  | Node_map2 : map2_set -> domain -> codomain -> map2_set -> nat -> map2_set.

  Definition height (m : map2_set) : nat :=
    match m with
      Empty_map2 => O
    | (Node_map2 _ _ _ _ h) => h
    end.
 
  Inductive map2_bounded_down (bound : domain) : map2_set -> Prop :=
    Empty_map2_bounded_down : (map2_bounded_down bound Empty_map2)
  | Node_map2_bounded_down :
      forall l : map2_set,
      forall d : domain,
      forall c : codomain,
      forall r : map2_set,
      forall h : nat,
      (map2_bounded_down bound l) ->
      (domain_le bound d) -> ~(domain_le d bound) ->
      (map2_bounded_down bound r) ->
      (map2_bounded_down bound (Node_map2 l d c r h)).

  Inductive map2_bounded_up (bound : domain) : map2_set -> Prop :=
    Empty_map2_bounded_up : (map2_bounded_up bound Empty_map2)
  | Node_map2_bounded_up :
      forall l : map2_set,
      forall d : domain,
      forall c : codomain,
      forall r : map2_set,
      forall h : nat,
      (map2_bounded_up bound l) ->
      (domain_le d bound) -> ~(domain_le bound d) ->
      (map2_bounded_up bound r) ->
      (map2_bounded_up bound (Node_map2 l d c r h)).

  Inductive map2_wo : map2_set -> Prop :=
    Empty_map2_wo : (map2_wo Empty_map2)
  | Node_map2_wo :
      forall l : map2_set,
      forall d : domain,
      forall c : codomain,
      forall r : map2_set, 
      forall h : nat,
      (map2_bounded_up d l) ->
      (map2_bounded_down d r) ->
      h = (S (max (height l) (height r))) ->
      (map2_wo (Node_map2 l d c r h)).

  Hint Resolve
    Empty_map2_bounded_down
    Node_map2_bounded_down
    Empty_map2_bounded_up
    Node_map2_bounded_up
    Empty_map2_wo
    Node_map2_wo : map2.

  Lemma map2_bounded_down_trans :
    forall m : map2_set,
    forall bound1 bound2 : domain,
    (domain_le bound1 bound2) ->
    (map2_bounded_down bound2 m) ->
    (map2_bounded_down bound1 m).
  Proof.
    induction m; intros bound1 bound2 Horder Hbound; inversion_clear Hbound;
    eauto 10 with map2.
  Qed.
  Hint Resolve map2_bounded_down_trans : map2.

  Lemma map2_bounded_up_trans :
    forall m : map2_set,
    forall bound1 bound2 : domain,
    (domain_le bound1 bound2) ->
    (map2_bounded_up bound1 m) ->
    (map2_bounded_up bound2 m).
  Proof.
    induction m; intros bound1 bound2 Horder Hbound; inversion_clear Hbound;
    eauto 10 with map2.
  Qed.
  Hint Resolve map2_bounded_up_trans : map2.

  Definition t := { x : map2_set | (map2_wo x) }.

  Fixpoint map2_find (lookup : domain) (m : map2_set) {struct m} :
    (option codomain) :=
    match m with
      Empty_map2 => None
    | (Node_map2 l d c r h) =>
      match (compare lookup d) with
        LESS _ _ => map2_find lookup l
      | EQUAL _ _ => Some c
      | GREATER _ _ => map2_find lookup r
      end
    end.

  Definition find (lookup : domain) (m : t) :=
    map2_find lookup (proj1_sig m).
End Map2_set.

Hint Resolve
    Empty_map2_bounded_down
    Node_map2_bounded_down
    Empty_map2_bounded_up
    Node_map2_bounded_up
    Empty_map2_wo
    Node_map2_wo
    map2_bounded_up_trans
    map2_bounded_down_trans : map2.

Section Map1.
  Variable codomain1 codomain2 : Set.
  Variable f : codomain1 -> codomain2.

  Fixpoint map2_map1 (m : (map2_set codomain1))
    {struct m} : (map2_set codomain2) :=
    match m with
      Empty_map2 => (Empty_map2 codomain2)
    | (Node_map2 l d c r h) =>
         (Node_map2 codomain2
            (map2_map1 l)
            d (f c)
            (map2_map1 r)
            h)
    end.

  Lemma map2_map1_bounded_down :
    forall bound : domain,
    forall m : (map2_set codomain1),
    (map2_bounded_down codomain1 bound m) ->
    (map2_bounded_down codomain2 bound (map2_map1 m)).
  Proof.
    induction m; intro Hbound; inversion Hbound; simpl in |- *; auto with map2.
  Qed.

  Lemma map2_map1_bounded_up :
    forall bound : domain,
    forall m : (map2_set codomain1),
    (map2_bounded_up codomain1 bound m) ->
    (map2_bounded_up codomain2 bound (map2_map1 m)).
  Proof.
    induction m; intro Hbound; inversion Hbound; simpl in |- *; auto with map2.
  Qed.

  Hint Resolve map2_map1_bounded_down : map2.
  Hint Resolve map2_map1_bounded_up : map2.

  Lemma map2_map1_height :
    forall m : (map2_set codomain1),
    (height codomain2 (map2_map1 m)) = (height codomain1 m).
  Proof.
    induction m; auto with arith.
  Qed.

  Hint Rewrite map2_map1_height : map2.

  Lemma map2_map1_wo :
    forall m : (map2_set codomain1),
    (map2_wo codomain1 m) ->
    (map2_wo codomain2 (map2_map1 m)).
  Proof.
    induction m; intro Hwo; inversion Hwo; simpl in |- *; auto with map2.
    apply Node_map2_wo; auto with map2.
    rewrite map2_map1_height.
    rewrite map2_map1_height.
    assumption.
  Qed.

  Definition map1 : (t codomain1) -> (t codomain2).
  Proof.
    intro m.
    destruct m as (m, Hmwo).
    exists (map2_map1 m).
    apply map2_map1_wo; assumption.
  Defined.

  Definition func_map1 (mapf : domain -> (option codomain1))
    (lookup : domain) :=
    match mapf lookup with
    | None => None
    | Some x => Some (f x)
    end.

  Lemma map2_map1_correct :
    forall m : (map2_set codomain1),
    forall lookup : domain,
    (func_map1 (fun lookup1 => map2_find codomain1 lookup1 m)) lookup =
    map2_find codomain2 lookup (map2_map1 m).
  Proof.
    unfold func_map1 in |- *; induction m; simpl in |- *.
    auto.
    intro.
    compare_cases lookup d.
  Qed.

  Theorem map1_correct :
    forall m : (t codomain1),
    forall lookup : domain,
    (func_map1 (fun lookup1 => find codomain1 lookup1 m)) lookup =
    find codomain2 lookup (map1 m).
  Proof.
    destruct m; unfold find, map1 in |- *; simpl in |- *.
    apply map2_map1_correct.
  Qed.
End Map1.
